import logging
from typing import Optional, Type, Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Model
from django.db.models.base import ModelBase  # post_migrate may call with phony objects
from django.db.models.signals import post_delete, post_migrate
from django.utils.functional import cached_property

from ansible_base.rbac.managed import ManagedRoleConstructor, get_managed_role_constructors

"""
This will record the models that the RBAC system in this app will follow
Other apps should register models with this pattern

from ansible_base.utils.permission_registry import permission_registry

permission_registry.register(MyModel, AnotherModel)
"""

logger = logging.getLogger('ansible_base.rbac.permission_registry')


class PermissionRegistry:
    def __init__(self):
        self._registry = set()  # model registry
        self._name_to_model = dict()
        self._parent_fields = dict()
        self._managed_roles = dict()  # code-defined role definitions, managed=True
        self.apps_ready = False
        self._tracked_relationships = set()
        self._trackers = dict()

    def register(self, *args: Type[Model], parent_field_name: Optional[str] = 'organization'):
        if self.apps_ready:
            raise RuntimeError('Cannot register model to permission_registry after apps are ready')
        for cls in args:
            if cls not in self._registry:
                self._registry.add(cls)
                model_name = cls._meta.model_name
                if model_name in self._name_to_model:
                    raise RuntimeError(f'Two models registered with same name {model_name}')
                self._name_to_model[model_name] = cls
                if model_name != 'organization':
                    self._parent_fields[model_name] = parent_field_name
            else:
                logger.debug(f'Model {cls._meta.model_name} registered to permission registry more than once')

    def track_relationship(self, cls, relationship, role_name):
        self._tracked_relationships.add((cls, relationship, role_name))

    def get_parent_model(self, model) -> Optional[type]:
        model = self._name_to_model[model._meta.model_name]
        parent_field_name = self.get_parent_fd_name(model)
        if parent_field_name is None:
            return None
        return model._meta.get_field(parent_field_name).related_model

    def get_parent_fd_name(self, model) -> Optional[str]:
        return self._parent_fields.get(model._meta.model_name)

    def get_child_models(self, parent_model, seen=None) -> list[tuple[str, Type[Model]]]:
        """Returns child models and the filter relationship to the parent

        This is used for rebuilding RoleEvaluation entries.
        For the given parent model like organization, this returns a list of tuples that contains
         - path like "parent__organization" in Model.objects.filter(parent__organization=organization)
         - the model class which is a child resource of the parent model
        """
        if not seen:
            seen = set()
        child_filters = []
        parent_model_name = parent_model._meta.model_name
        for model_name, parent_field_name in self._parent_fields.items():
            if parent_field_name is None:
                continue
            child_model = self._name_to_model[model_name]
            this_parent_name = child_model._meta.get_field(parent_field_name).related_model._meta.model_name
            if this_parent_name == parent_model_name:
                if model_name in seen:
                    continue
                seen.add(model_name)

                child_filters.append((parent_field_name, child_model))
                for next_parent_filter, grandchild_model in self.get_child_models(child_model, seen=seen):
                    child_filters.append((f'{next_parent_filter}__{parent_field_name}', grandchild_model))
        return child_filters

    def get_resource_prefix(self, cls: Type[Model]) -> str:
        """For a given model class, give the prefix like shared, of API naming like shared.team"""
        if registry := self.get_resource_registry():
            # duplicates logic in ansible_base/resource_registry/apps.py
            try:
                resource_config = registry.get_config_for_model(cls)
                if resource_config.managed_serializer:
                    return "shared"  # shared model
            except KeyError:
                pass  # unregistered model

            # Fallback for unregistered and non-shared models
            return registry.api_config.service_type
        else:
            return 'local'

    def get_resource_registry(self):
        if 'ansible_base.resource_registry' not in settings.INSTALLED_APPS:
            return None

        from ansible_base.resource_registry.registry import get_registry

        return get_registry()

    def get_managed_role_constructor(self, shortname: str) -> Optional[ManagedRoleConstructor]:
        return self._managed_roles.get(shortname)

    def get_managed_role_constructor_by_name(self, name: str) -> Optional[ManagedRoleConstructor]:
        for managed_role in self._managed_roles.values():
            if managed_role.name == name:
                return managed_role

    def register_managed_role_constructor(self, shortname: str, managed_role: ManagedRoleConstructor) -> None:
        """Add the given managed role to the managed role registry"""
        self._managed_roles[shortname] = managed_role

    def register_managed_role_constructors(self) -> None:
        """Adds the data in setting ANSIBLE_BASE_MANAGED_ROLE_REGISTRY to the managed role registry"""
        managed_defs = get_managed_role_constructors(self.apps, settings.ANSIBLE_BASE_MANAGED_ROLE_REGISTRY)
        for shortname, constructor in managed_defs.items():
            self.register_managed_role_constructor(shortname, constructor)

    def create_managed_roles(self, apps) -> list[tuple[Model, bool]]:
        """Safe-ish method to create managed roles inside of a migration

        Returns a list with all the managed RoleDefinition objects and whether they were created
        in case you have to make decisions based on that"""
        if not self.apps_ready:
            raise RuntimeError('Cannot create managed roles before apps are ready')
        ret = []
        for managed_role in self._managed_roles.values():
            rd, created = managed_role.get_or_create(apps)
            ret.append((rd, created))
        return ret

    def call_when_apps_ready(self, apps, app_config) -> None:
        from ansible_base.rbac import triggers
        from ansible_base.rbac.evaluations import bound_has_obj_perm, bound_singleton_permissions, connect_rbac_methods
        from ansible_base.rbac.management import create_dab_permissions

        self.apps = apps

        # Finish registering models
        if self.team_model not in self._registry:
            self.register(self.team_model)

        for model_name, kwargs in settings.ANSIBLE_BASE_RBAC_MODEL_REGISTRY.items():
            model = apps.get_model(model_name)
            if model not in self._registry:
                self.register(model, **kwargs)

        # This will lock-down the registry, raising an error for any other registrations
        self.apps_ready = True

        # Do no specify sender for create_dab_permissions, because that is passed as app_config
        # and we want to create permissions for external apps, not the dab_rbac app
        post_migrate.connect(
            create_dab_permissions,
            dispatch_uid="ansible_base.rbac.management.create_dab_permissions",
        )
        post_migrate.connect(
            triggers.post_migration_rbac_setup,
            sender=app_config,
            dispatch_uid="ansible_base.rbac.triggers.post_migration_rbac_setup",
        )

        self.user_model.add_to_class('has_obj_perm', bound_has_obj_perm)
        self.user_model.add_to_class('singleton_permissions', bound_singleton_permissions)
        post_delete.connect(triggers.rbac_post_user_delete, sender=self.user_model, dispatch_uid='permission-registry-user-delete')

        for cls in self._registry:
            triggers.connect_rbac_signals(cls)
            connect_rbac_methods(cls)

        for cls, relationship, role_name in self._tracked_relationships:
            if role_name in self._trackers:
                tracker = self._trackers[role_name]
            else:
                tracker = triggers.TrackedRelationship(cls, role_name)
                self._trackers[role_name] = tracker
            tracker.initialize(relationship)

        self.register_managed_role_constructors()

    @property
    def team_model(self):
        return self.apps.get_model(settings.ANSIBLE_BASE_TEAM_MODEL)

    @cached_property
    def team_ct_id(self):
        return self.content_type_model.objects.get_for_model(self.team_model).id

    @property
    def user_model(self):
        return get_user_model()

    @property
    def content_type_model(self):
        return self.apps.get_model('contenttypes.ContentType')

    @cached_property
    def org_ct_id(self):
        team_parent_model = self.get_parent_model(self.team_model)
        return self.content_type_model.objects.get_for_model(team_parent_model).id

    @property
    def permission_qs(self):
        """Return a queryset of the permission model restricted to the RBAC-tracked models

        Note that this should not be necessary, since the post_migrate signal for DABPermission
        will only create entries for registered models.
        However, removing permission entries after a model definition changes is still unsolved
        and this is already problematic for auth.Permission.
        """
        all_cts = self.content_type_model.objects.get_for_models(*self.all_registered_models)
        return self.apps.get_model('dab_rbac.DABPermission').objects.filter(content_type__in=all_cts.values())

    @property
    def team_permission(self):
        return f'member_{self.team_model._meta.model_name}'

    @property
    def all_registered_models(self):
        return list(self._registry)

    def is_registered(self, obj: Union[ModelBase, Model]) -> bool:
        """Tells if the given object or class is a type tracked by DAB RBAC"""
        return any((obj._meta.model_name == cls._meta.model_name and obj._meta.app_label == cls._meta.app_label) for cls in self._registry)

    def get_model_by_name(self, model_name: str) -> Optional[Type[Model]]:
        """Returns class with given model_name if registered, returns None otherwise"""
        for cls in self._registry:
            if model_name == cls._meta.model_name:
                return cls
        return None


permission_registry = PermissionRegistry()



# Generated by Django 4.2.16 on 2024-11-21 11:14

import ansible_base.oauth2_provider.models.access_token
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_oauth2_provider', '0005_hash_existing_tokens'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='scope',
            field=models.CharField(default='write', help_text="Allowed scopes, further restricts user permissions. Must be a simple space-separated string with allowed scopes ['read', 'write'].", max_length=32, validators=[ansible_base.oauth2_provider.models.access_token.validate_scope]),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='user',
            field=models.ForeignKey(blank=True, help_text='The user representing the token owner.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='access_tokens', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='name',
            field=models.CharField(help_text='The name of this resource.', max_length=512),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
    ]



import logging
import multiprocessing
import random
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

from django.core import cache as django_cache
from django.core.cache.backends.base import BaseCache

from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.cache.fallback_cache')

DEFAULT_TIMEOUT = None
PRIMARY_CACHE = 'primary'
FALLBACK_CACHE = 'fallback'


class DABCacheWithFallback(BaseCache):
    _instance = None
    _primary_cache = None
    _fallback_cache = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(DABCacheWithFallback, cls).__new__(cls)
            cls.__initialized = False
        return cls._instance

    def __init__(self, location, params):
        if self.__initialized:
            return
        BaseCache.__init__(self, params)

        self._primary_cache = django_cache.caches.create_connection(PRIMARY_CACHE)
        self._fallback_cache = django_cache.caches.create_connection(FALLBACK_CACHE)
        self._temp_path = get_setting('ANSIBLE_BASE_FALLBACK_CACHE_FILE_PATH', tempfile.gettempdir())
        self._temp_file = Path().joinpath(self._temp_path, 'gw_primary_cache_failed')
        self.thread_pool = ThreadPoolExecutor()

        if self._temp_file.exists():
            self._temp_file.unlink()

        self.__initialized = True

    def get_active_cache(self):
        return FALLBACK_CACHE if self._temp_file.exists() else PRIMARY_CACHE

    # Main cache interface
    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        return self._op_with_fallback("add", key, value, timeout=timeout, version=version)

    def get(self, key, default=None, version=None):
        return self._op_with_fallback("get", key, default=default, version=version)

    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None, client=None):
        return self._op_with_fallback("set", key, value, timeout=timeout)

    def delete(self, key, version=None):
        return self._op_with_fallback("delete", key, version=version)

    def clear(self):
        return self._op_with_fallback("clear")

    def _op_with_fallback(self, operation, *args, **kwargs):
        if self._temp_file.exists():
            response = getattr(self._fallback_cache, operation)(*args, **kwargs)
            self.start_recovery()
        else:
            try:
                response = getattr(self._primary_cache, operation)(*args, **kwargs)
                return response
            except Exception:
                with multiprocessing.Lock():
                    # Attempt to ensure one thread/process goes first
                    # dynamic settings especially are read in a batch very quickly
                    time.sleep(random.uniform(10, 100) / 100.0)
                    if not self._temp_file.exists():
                        logger.error("Primary cache unavailable, switching to fallback cache.")
                    self.create_temp_file()
                    # _temp_file.touch()
                response = getattr(self._fallback_cache, operation)(*args, **kwargs)

        return response

    def start_recovery(self):
        with multiprocessing.Lock():
            # Set single process/thread to do the recovery, but time out in case it dies
            recoverer = self._fallback_cache.get_or_set('RECOVERY_THREAD_ID', id(self), timeout=60)
            if recoverer == id(self):
                rip = self._fallback_cache.get('RECOVERY_IN_PROGRESS', False)
                if not rip:
                    self._fallback_cache.set('RECOVERY_IN_PROGRESS', True, timeout=60)
                    self.thread_pool.submit(self.check_primary_cache)

    def check_primary_cache(self):
        try:
            self._primary_cache.get('up_test')
            with multiprocessing.Lock():
                if self._temp_file.exists():
                    logger.warning("Primary cache recovered, clearing and resuming use.")
                    # Clear the primary cache
                    self._primary_cache.clear()
                    # Clear the backup cache just incase we need to fall back again (don't want it out of sync)
                    self._fallback_cache.clear()
                    self._temp_file.unlink()
        except Exception:
            pass
        finally:
            self._fallback_cache.delete('RECOVERY_IN_PROGRESS')

    def create_temp_file(self):
        self._temp_file.touch()
        self._temp_file.chmod(mode=0o660)



import logging

from django.conf import settings
from django.contrib.auth.backends import ModelBackend

PREFIX = getattr(settings, "RENAMED_USERNAME_PREFIX", None)
logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.github_enterprise_team')


class PrefixedUserAuthenticationMixin:
    def authenticate(self, request, **kwargs):
        if not PREFIX:
            return None
        if username := kwargs.get("username", None):
            if not username.startswith(PREFIX):
                kwargs["username"] = PREFIX + username
                return super().authenticate(request, **kwargs)


class PrefixedUserAuthBackend(PrefixedUserAuthenticationMixin, ModelBackend):
    pass



import logging
from typing import Optional, Tuple, Type

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext as _

from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.lib.utils.create_system_user')

"""
These functions are in its own file because it is loaded during migrations so it has no access to models.
"""


def create_system_user(user_model: Type[models.Model]) -> models.Model:  # Note: We can't load models here so we can typecast to anything better than Model
    from ansible_base.lib.abstract_models.user import AbstractDABUser

    #
    # Creating the system user is a little tricky because we need to reference ourselves
    #
    if not user_model:
        logger.error("You must pass a user_model to create_system_user")
        return None

    # First create a User object for the system user

    # If we use subclass of AbstractDABUser ensure we use manager for unfiltered queryset
    user_manager = user_model.all_objects if issubclass(user_model, AbstractDABUser) else user_model.objects
    system_user, created = user_manager.get_or_create(username=get_system_username()[0])

    if created:
        logger.info(f"Created system user {system_user.username}")
    else:
        logger.debug("System user is already created")
        return system_user

    system_user.is_active = False
    system_user.set_unusable_password()
    system_user.created_by = system_user
    system_user.modified_by = system_user
    update_fields = ['is_active', 'password', 'created_by', 'modified_by']

    # If our model has a managed flag set it to true
    if hasattr(user_model, 'managed'):
        system_user.managed = True
        update_fields.append('managed')

    system_user.save(update_fields=update_fields)

    return system_user


def get_system_username() -> Tuple[Optional[str], str]:
    # Returns (system_username, setting_name)
    setting_name = 'SYSTEM_USERNAME'
    value = get_setting(setting_name)
    if value is None:
        return None, setting_name

    if isinstance(value, str):
        return str(value), setting_name

    logger.error(f"Expected get_setting to return a string for {setting_name}, got {type(value)}")
    raise ImproperlyConfigured(_("Setting %(setting_name)s needs to be a string not a %(type)s") % {'setting_name': setting_name, 'type': type(value)})



import logging

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _
from rest_framework.serializers import ValidationError

from ansible_base.authentication.models import AuthenticatorMap
from ansible_base.authentication.utils.trigger_definition import TRIGGER_DEFINITION
from ansible_base.lib.serializers.common import NamedCommonModelSerializer
from ansible_base.lib.utils.auth import get_organization_model, get_team_model
from ansible_base.lib.utils.string import is_empty

logger = logging.getLogger('ansible_base.authentication.serializers.authenticator_map')


class AuthenticatorMapSerializer(NamedCommonModelSerializer):
    class Meta:
        model = AuthenticatorMap
        fields = NamedCommonModelSerializer.Meta.fields + ['authenticator', 'map_type', 'role', 'organization', 'team', 'revoke', 'triggers', 'order']

    def validate(self, data) -> dict:
        errors = {}
        errors.update(self.validate_trigger_data(data))

        map_type = data.get('map_type', None)
        team = data.get('team', None)
        org = data.get('organization', None)
        role = data.get('role', None)

        if map_type == 'team' and is_empty(team):
            errors["team"] = _("You must specify a team with the selected map type")
        if map_type in ['team', 'organization'] and is_empty(org):
            errors["organization"] = _("You must specify an organization with the selected map type")
        if map_type in ['team', 'organization', 'role'] and is_empty(role):
            errors["role"] = _("You must specify a role with the selected map type")
        if map_type in ['allow', 'is_superuser'] and not is_empty(role):
            errors["role"] = _("You cannot specify role with the selected map type")

        if role:
            errors.update(self.validate_role_data(map_type, role, org, team))

        if errors:
            raise ValidationError(errors)
        return data

    @staticmethod
    def _is_rbac_installed():
        return 'ansible_base.rbac' in settings.INSTALLED_APPS

    def validate_role_data(self, map_type, role, org, team):
        errors = {}

        # Validation is possible only if RBAC is installed
        if not self._is_rbac_installed():
            logger.warning(_("You specified a role without RBAC installed "))
            return errors

        from ansible_base.rbac.models import RoleDefinition

        try:
            rbac_role = RoleDefinition.objects.get(name=role)
            is_system_role = rbac_role.content_type is None

            # system role is allowed for map type == role without further conditions
            if is_system_role and map_type == 'role':
                return errors

            if is_system_role:
                is_org_role, is_team_role = False, False
            else:
                model_class = rbac_role.content_type.model_class()
                is_org_role = issubclass(model_class, get_organization_model())
                is_team_role = issubclass(model_class, get_team_model())

            # role type and map type must correspond
            if map_type == 'organization' and not is_org_role:
                errors['role'] = _("For an organization map type you must specify an organization based role")

            if map_type == 'team' and not is_team_role:
                errors['role'] = _("For a team map type you must specify a team based role")

            # org/team role needs organization field
            if (is_org_role or is_team_role) and is_empty(org):
                errors["organization"] = _("You must specify an organization with the selected role")

            # team role needs team field
            if is_team_role and is_empty(team):
                errors["team"] = _("You must specify a team with the selected role")

        except ObjectDoesNotExist:
            errors['role'] = _("RoleDefinition {role} doesn't exist").format(role=role)

        return errors

    def validate_trigger_data(self, data):
        errors = {}
        request = self.context.get('request', None)
        if 'triggers' not in data or not data['triggers']:
            if not request or (request.method != 'PATCH'):
                errors["triggers"] = "Triggers must be a valid dict"
        else:
            errors.update(self._validate_trigger_data(data['triggers'], TRIGGER_DEFINITION, 'triggers'))
        return errors

    def _validate_trigger_data(self, triggers: dict, definition, error_prefix: str) -> dict:
        """
        Examples of valid data:
        - {triggers: {'groups': {'has_or': ['aaa', 'bbb'], 'has_and': ['ccc']}}}
        - {triggers: {'always': {}}}
        - {triggers: {'never': {}}}
        - {triggers: {'attributes': {'join_condition': "and",
                                   'some_attr1': {'contains': "some_str"},
                                   'some_attr2': {'ends_with': "some_str"}}}}
        """
        errors = {}

        # Validate only valid items
        for trigger_type in triggers.keys():
            type_definition = definition.get(trigger_type, definition.get('*', None))
            if not type_definition:
                errors[f'{error_prefix}.{trigger_type}'] = f"Invalid, can only be one of: {', '.join(definition.keys())}"
                continue

            # Validate the type we got is what we expect
            if not isinstance(triggers[trigger_type], type(type_definition['type'])):
                errors[f'{error_prefix}.{trigger_type}'] = f"Expected {type(type_definition['type']).__name__} but got {type(triggers[trigger_type]).__name__}"
                continue

            if isinstance(triggers[trigger_type], dict):
                errors.update(self._validate_trigger_data(triggers[trigger_type], type_definition['keys'], f'{error_prefix}.{trigger_type}'))
            elif isinstance(triggers[trigger_type], str):
                if 'choices' in type_definition:
                    if triggers[trigger_type] not in type_definition['choices']:
                        errors[f'{error_prefix}.{trigger_type}'] = f"Invalid, choices can only be one of: {', '.join(type_definition['choices'])}"
            elif isinstance(triggers[trigger_type], list):
                if 'contents' in type_definition:
                    for item in triggers[trigger_type]:
                        if not isinstance(item, type(type_definition['contents'])):
                            errors[f'{error_prefix}.{trigger_type}.{item}'] = (
                                f"Invalid, must be of type {type(type_definition['contents']).__name__}, got {type(item)}"
                            )

        return errors



import hashlib

import oauth2_provider.models as oauth2_models
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import connection, models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from oauthlib import oauth2

from ansible_base.lib.abstract_models.common import CommonModel
from ansible_base.lib.utils.hashing import hash_string
from ansible_base.lib.utils.models import prevent_search
from ansible_base.lib.utils.settings import get_setting
from ansible_base.oauth2_provider.utils import is_external_account

SCOPES = ['read', 'write']


def validate_scope(value):
    given_scopes = value.split(' ')
    if not given_scopes:
        raise ValidationError(_('Scope must be a simple space-separated string with allowed scopes: %(scopes)s') % {'scopes': ', '.join(SCOPES)})
    for scope in given_scopes:
        if scope not in SCOPES:
            raise ValidationError(_('Invalid scope: %(scope)s. Must be one of: %(scopes)s') % {'scope': scope, 'scopes': ', '.join(SCOPES)})


activitystream = object
if 'ansible_base.activitystream' in settings.INSTALLED_APPS:
    from ansible_base.activitystream.models import AuditableModel

    activitystream = AuditableModel


class OAuth2AccessToken(CommonModel, oauth2_models.AbstractAccessToken, activitystream):
    router_basename = 'token'
    ignore_relations = ['refresh_token']
    activity_stream_excluded_field_names = ['last_used']

    class Meta(oauth2_models.AbstractAccessToken.Meta):
        verbose_name = _('access token')
        ordering = ('id',)
        swappable = "OAUTH2_PROVIDER_ACCESS_TOKEN_MODEL"

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="access_tokens",
        help_text=_('The user representing the token owner.'),
    )
    # Overriding to set related_name
    application = models.ForeignKey(
        settings.OAUTH2_PROVIDER_APPLICATION_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name='access_tokens',
        help_text=_('The related application. If None, this is a user token instead of an application token.'),
    )
    description = models.TextField(default='', blank=True, help_text=_('A description for this token.'))
    last_used = models.DateTimeField(null=True, default=None, editable=False, help_text=_('A timestamp of when this token was last used.'))
    scope = models.CharField(
        default='write',
        max_length=32,
        help_text=_("Allowed scopes, further restricts user permissions. Must be a simple space-separated string with allowed scopes ['read', 'write']."),
        validators=[validate_scope],
    )
    token = prevent_search(models.CharField(max_length=255, unique=True, help_text=_("The generated token value.")))
    updated = None  # Tracked in CommonModel with 'modified', no need for this

    def is_valid(self, scopes=None):
        valid = super(OAuth2AccessToken, self).is_valid(scopes)
        if valid:
            self.last_used = now()

            def _update_last_used():
                if OAuth2AccessToken.objects.filter(pk=self.pk).exists():
                    self.save(update_fields=['last_used'])

            connection.on_commit(_update_last_used)
        return valid

    def validate_external_users(self):
        if self.user and get_setting('ALLOW_OAUTH2_FOR_EXTERNAL_USERS') is False:
            external_account = is_external_account(self.user)
            if external_account:
                raise oauth2.AccessDeniedError(
                    _('OAuth2 Tokens cannot be created by users associated with an external authentication provider (%(authenticator)s)')
                    % {'authenticator': external_account.name}
                )

    def save(self, *args, **kwargs):
        if not self.pk:
            self.validate_external_users()
            self.token = hash_string(self.token, hasher=hashlib.sha256, algo="sha256")
        super().save(*args, **kwargs)



# Python
import logging

from ansible_base.jwt_consumer.common.auth import JWTAuthentication

logger = logging.getLogger('ansible_base.jwt_consumer.awx.auth')


class AwxJWTAuthentication(JWTAuthentication):
    use_rbac_permissions = True
